/*==============================================================================


   Purpose:    low level routines for Pentiums/Linux


   Author:     Rudolf Berrendorf
               University of Applied Sciences Rhein-Sieg
               Computer Science Department
               53754 Sankt Augustin
               rudolf.berrendorf@fh-rhein-sieg.de


   last mod:   01.09.2000


   DISCLAIMER

                  PCL - The Performance Counter Library


   Copyright (C) 2000
   Forschungszentrum Juelich GmbH, Federal Republic of Germany.
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are met:

   Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

     - Redistributions of source code must retain the above copyright notice,
       this list of conditions and the following disclaimer.

     - Redistributions in binary form must reproduce the above copyright
       notice, this list of conditions and the following disclaimer in the
       documentation and/or other materials provided with the distribution.

     - Any publications that result from the use of this software shall
       reasonably refer to the Research Centre's development.

     - All advertising materials mentioning features or use of this software
       must display the following acknowledgement:

           This product includes software developed by Forschungszentrum
           Juelich GmbH, Federal Republic of Germany.

     - Forschungszentrum Juelich GmbH is not obligated to provide the user with
       any support, consulting, training or assistance of any kind with regard
       to the use, operation and performance of this software or to provide
       the user with any updates, revisions or new versions.


   THIS SOFTWARE IS PROVIDED BY FORSCHUNGSZENTRUM JUELICH GMBH "AS IS" AND ANY
   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
   DISCLAIMED. IN NO EVENT SHALL FORSCHUNGSZENTRUM JUELICH GMBH BE LIABLE FOR
   ANY SPECIAL, DIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
   RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
   CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
   CONNECTION WITH THE ACCESS, USE OR PERFORMANCE OF THIS SOFTWARE.
  
==============================================================================*/

#include "../../defines.h"
#include "../../pcl.h"
#include "hw_driver_defs.h"
#include "../../pcl_hw_driver.h"
#include "../../hw_driver_global_defs.h"
#include "../../state_descr.h"

/*                                                                            */
/*----------------------------------------------------------------------------*/
/* constants                                                                  */

/* model specific registers for Pentium and Pentium MMX */
#define PENTIUM_MSR_CTLEVSEL 0x11   /* MSR for control and event select */
#define PENTIUM_MSR_CNT0     0x12   /* MSR for counter 0 */
#define PENTIUM_MSR_CNT1     0x13   /* MSR for counter 1 */

/* model specific registers for PentiumPro, Pentium II, Pentium III */
#define PENTIUM2_MSR_EVNTSEL0 0x186 /* MSR for control and event select ctr0 */
#define PENTIUM2_MSR_EVNTSEL1 0x187 /* MSR for control and event select ctr1 */
#define PENTIUM2_MSR_PERFCTR0 0xc1  /* MSR for counter 0 */
#define PENTIUM2_MSR_PERFCTR1 0xc2  /* MSR for counter 1 */

/*                                                                            */
/*----------------------------------------------------------------------------*/
/* local variables                                                            */


/* invalid entry-data if event is not available */
#define PCL_INVALID    0, {{{-1,-1,-1},{0,0,0}}}

static const PCL_event_descr_t event_table_pentium[PCL_MAX_EVENT] =
{
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1CACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1CACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1CACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1CACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1CACHE_MISS */
  {PCL_AVAIL,NOT_COMP,{{{0x00,0x00,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_L1DCACHE_READ */
  {PCL_AVAIL,NOT_COMP,{{{0x01,0x01,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_L1DCACHE_WRITE */
  {PCL_AVAIL,COMPOSED,{{{0x00,0x01,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_L1DCACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1DCACHE_HIT */
  {PCL_AVAIL,NOT_COMP,{{{0x29,0x29,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_L1DCACHE_MISS */
  {PCL_AVAIL,NOT_COMP,{{{0x0C,0x0C,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_L1ICACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1ICACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1ICACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1ICACHE_HIT */
  {PCL_AVAIL,NOT_COMP,{{{0x0E,0x0E,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_L1ICACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2CACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2CACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2CACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2CACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2CACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2DCACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2DCACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2DCACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2DCACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2DCACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2ICACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2ICACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2ICACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2ICACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2ICACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_TLB_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_TLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_ITLB_HIT */
  {PCL_AVAIL,NOT_COMP,{{{0x0D,0x0D,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_ITLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_DTLB_HIT */
  {PCL_AVAIL,NOT_COMP,{{{0x02,0x02,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_DTLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_CYCLES */
  /* special case with dummy entry */
  {PCL_AVAIL,NOT_COMP,{{{-1,-1,0x00},{-1,-1,OP_ADD}}}},  /* PCL_ELAPSED_CYCLES */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_INTEGER_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{0x22,0x22,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_FP_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_LOAD_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_STORE_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{0x28,0x28,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_LOADSTORE_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{0x16,0x16,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_JUMP_SUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_JUMP_UNSUCCESS */
  {PCL_AVAIL,NOT_COMP,{{{0x12,0x12,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_JUMP */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_ATOMIC_SUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_ATOMIC_UNSUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_ATOMIC */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_STALL_INTEGER */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_STALL_FP */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_STALL_JUMP */
  {PCL_AVAIL,NOT_COMP,{{{0x1A,0x1A,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_STALL_LOAD */
  {PCL_AVAIL,NOT_COMP,{{{0x19,0x19,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_STALL_STORE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_STALL */
  {PCL_UNAVAIL, PCL_INVALID},          /* PCL_MFLOPS */
  {PCL_UNAVAIL, PCL_INVALID},          /* PCL_IPC */
  {PCL_AVAIL,COMPOSED,{{{0x29,0x28,-1},{OP_FP_2|OP_FP_STORE,OP_FP_2|OP_FP_DIV,-1}}}}, /* PCL_L1DCACHE_MISSRATE */
  {PCL_UNAVAIL, PCL_INVALID},          /* PCL_L2DCACHE_MISSRATE */
  {PCL_AVAIL,COMPOSED,{{{0x28,0x22,-1},{OP_FP_2|OP_FP_STORE,OP_FP_2|OP_FP_DIV,-1}}}} /* PCL_MEM_FP_RATIO */
};


static const PCL_event_descr_t event_table_pentium_mmx[PCL_MAX_EVENT] =
{
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1CACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1CACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1CACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1CACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1CACHE_MISS */
  {PCL_AVAIL,NOT_COMP,{{{0x00,0x00,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_L1DCACHE_READ */
  {PCL_AVAIL,NOT_COMP,{{{0x01,0x01,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_L1DCACHE_WRITE */
  {PCL_AVAIL,COMPOSED,{{{0x00,0x01,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_L1DCACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1DCACHE_HIT */
  {PCL_AVAIL,NOT_COMP,{{{0x29,0x29,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_L1DCACHE_MISS */
  {PCL_AVAIL,NOT_COMP,{{{0x0C,0x0C,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_L1ICACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1ICACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1ICACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1ICACHE_HIT */
  {PCL_AVAIL,NOT_COMP,{{{0x0E,0x0E,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_L1ICACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2CACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2CACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2CACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2CACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2CACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2DCACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2DCACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2DCACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2DCACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2DCACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2ICACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2ICACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2ICACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2ICACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2ICACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_TLB_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_TLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_ITLB_HIT */
  {PCL_AVAIL,NOT_COMP,{{{0x0D,0x0D,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_ITLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_DTLB_HIT */
  {PCL_AVAIL,NOT_COMP,{{{0x02,0x02,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_DTLB_MISS */
  {PCL_AVAIL,NOT_COMP,{{{0x30,-1,-1},{OP_ADD,-1,-1}}}}, /* PCL_CYCLES */
  /* special case with dummy entry */
  {PCL_AVAIL,NOT_COMP,{{{-1,-1,0x00},{-1,-1,OP_ADD}}}},  /* PCL_ELAPSED_CYCLES */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_INTEGER_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{0x22,0x22,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_FP_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_LOAD_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_STORE_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{0x28,0x28,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_LOADSTORE_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{0x16,0x16,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{-1,0x31,-1},{-1,OP_ADD,-1}}}}, /* PCL_JUMP_SUCCESS */
  {PCL_AVAIL,COMPOSED,{{{0x12,0x31,-1},{OP_ADD,OP_SUB,-1}}}}, /* PCL_JUMP_UNSUCCESS */
  {PCL_AVAIL,NOT_COMP,{{{0x12,0x12,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_JUMP */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_ATOMIC_SUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_ATOMIC_UNSUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_ATOMIC */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_STALL_INTEGER */
  {PCL_AVAIL,NOT_COMP,{{{0x32,-1,-1},{OP_ADD,-1,-1}}}},  /* PCL_STALL_FP */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_STALL_JUMP */
  {PCL_AVAIL,NOT_COMP,{{{0x1A,0x1A,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_STALL_LOAD */
  {PCL_AVAIL,NOT_COMP,{{{0x19,0x19,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_STALL_STORE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_STALL */
  {PCL_AVAIL,COMPOSED,{{{0x30,0x22,-1},{OP_FP_1|OP_FP_2|OP_FP_REZ_MHZ,OP_FP_2|OP_FP_MUL,-1}}}}, /* PCL_MFLOPS */
  {PCL_AVAIL,COMPOSED,{{{0x30,0x16,-1},{OP_FP_1|OP_FP_2|OP_FP_REZ,OP_FP_2|OP_FP_MUL,-1}}}}, /* PCL_IPC */
  {PCL_AVAIL,COMPOSED,{{{0x29,0x28,-1},{OP_FP_2|OP_FP_STORE,OP_FP_2|OP_FP_DIV,-1}}}}, /* PCL_L1DCACHE_MISSRATE */
  {PCL_UNAVAIL, PCL_INVALID},          /* PCL_L2DCACHE_MISSRATE */
  {PCL_AVAIL,COMPOSED,{{{0x28,0x22,-1},{OP_FP_2|OP_FP_STORE,OP_FP_2|OP_FP_DIV,-1}}}} /* PCL_MEM_FP_RATIO */
};


static const PCL_event_descr_t event_table_pentium2[PCL_MAX_EVENT] =
{
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1CACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1CACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1CACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1CACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1CACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1DCACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1DCACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1DCACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1DCACHE_HIT */
  {PCL_AVAIL,NOT_COMP,{{{0x45,0x45,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_L1DCACHE_MISS */
  {PCL_AVAIL,NOT_COMP,{{{0x80,0x80,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_L1ICACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1ICACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1ICACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1ICACHE_HIT */
  {PCL_AVAIL,NOT_COMP,{{{0x81,0x81,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_L1ICACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2CACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2CACHE_WRITE */
  {PCL_AVAIL,NOT_COMP,{{{0x2E,0x2E,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_L2CACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2CACHE_HIT */
  {PCL_AVAIL,NOT_COMP,{{{0x24,0x24,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_L2CACHE_MISS */
  {PCL_AVAIL,NOT_COMP,{{{0x29,0x29,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_L2DCACHE_READ */
  {PCL_AVAIL,NOT_COMP,{{{0x2A,0x2A,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_L2DCACHE_WRITE */
  {PCL_AVAIL,COMPOSED,{{{0x29,0x2A,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_L2DCACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2DCACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2DCACHE_MISS */
  {PCL_AVAIL,NOT_COMP,{{{0x28,0x28,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_L2ICACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2ICACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2ICACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2ICACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2ICACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_TLB_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_TLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_ITLB_HIT */
  {PCL_AVAIL,NOT_COMP,{{{0x85,0x85,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_ITLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_DTLB_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_DTLB_MISS */
  {PCL_AVAIL,NOT_COMP,{{{0x79,0x79,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_CYCLES */
  /* special case with dummy entry */
  {PCL_AVAIL,NOT_COMP,{{{-1,-1,0x00},{-1,-1,OP_ADD}}}},  /* PCL_ELAPSED_CYCLES */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_INTEGER_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{0xC1,-1,-1},{OP_ADD,-1,-1}}}},  /* PCL_FP_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_LOAD_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_STORE_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{0x43,0x43,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_LOADSTORE_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{0xC0,0xC0,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{0xC9,0xC9,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_JUMP_SUCCESS */
  {PCL_AVAIL,NOT_COMP,{{{0xC5,0xC5,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_JUMP_UNSUCCESS */
  {PCL_AVAIL,NOT_COMP,{{{0xC4,0xC4,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_JUMP */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_ATOMIC_SUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_ATOMIC_UNSUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_ATOMIC */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_STALL_INTEGER */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_STALL_FP */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_STALL_JUMP */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_STALL_LOAD */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_STALL_STORE */
  {PCL_AVAIL,NOT_COMP,{{{0xA2,0xA2,-1},{OP_ADD,OP_ADD,-1}}}},  /* PCL_STALL */
  {PCL_AVAIL,COMPOSED,{{{0xC1,0x79,-1},{OP_FP_2|OP_FP_STORE,OP_FP_2|OP_FP_DIV_MHZ,-1}}}}, /* PCL_MFLOPS */
  {PCL_AVAIL,COMPOSED,{{{0xC0,0x79,-1},{OP_FP_2|OP_FP_STORE,OP_FP_2|OP_FP_DIV,-1}}}}, /* PCL_IPC */
  {PCL_AVAIL,COMPOSED,{{{0x45,0x43,-1},{OP_FP_2|OP_FP_STORE,OP_FP_2|OP_FP_DIV,-1}}}}, /* PCL_L1DCACHE_MISSRATE */
  {PCL_UNAVAIL, PCL_INVALID},          /* PCL_L2DCACHE_MISSRATE */
  {PCL_UNAVAIL, PCL_INVALID}           /* PCL_MEM_FP_RATIO */
};

/*                                                                            */
/*============================================================================*/
/* low-level stuff                                                            */


/* fast assembler code to access the cycle counter (TSC) */
#define rdtsc(valptr) asm volatile ("rdtsc" : \
"=a"(*(unsigned int *)valptr), "=d"(*(((unsigned int *)valptr)+1)))


/*                                                                            */
/*============================================================================*/
/* We determine if the processor supports performance counters (special MSR)
   and what events are supported. Therefore we use the CPUID instruction to
   test certain features.
   Sets "PCL_event_table" accordingly to get the right register settings.
   Returns PCL_SUCCESS if processor type is supported.
*/

int
PCL_driver_open(
		PCL_DESCR_TYPE d /* I: descriptor */
		)

{
  unsigned int signature, feature_flags, family;
  PCL_state_descr *descr;
  HW_state_descr *hw_descr;


  /* get state descriptor */
  descr = d;
  hw_descr = &descr->hw_state_descr;

  /* initialize global state variables */
  descr->allowed_count_mode = PCL_MODE_USER_SYSTEM;

  /* get all processor type information with the CPUID instruction
     Calling CPUID with EAX=1 returns in EAX version information (type, family,
     model, stepping ID) and in EDX feature information.
     The information in EAX is:
     bit  0-3 : stepping
     bit  4-7 : model
     bit  8-11: family
                0100 = i486
                0101 = Pentium
                0110 = Pentium Pro or Pentium II or Pentium III
     bit 12-13: processor type
     The information in EDX is (only relevant information shown here):
     bit 4 : Time Stamp Counter supported
     bit 5 : Model Specific Registers implemented with RDMSR, WRMSR instructions
     bit 23: MMX supported
   */
  asm("cpuid"
      : "=a"(signature), "=d"(feature_flags)
      : "a"(1)
      : "%ebx", "%ecx");

  /* check for existence of time stamp counter (TSC) */
  if((((feature_flags >> 4) & 0x1) != 1)
     /* and check for existence of model specific registers (MSR) */
     || (((feature_flags >> 5) & 0x1) != 1))
    /* something old without support for performance counters */
    return PCL_NOT_SUPPORTED;

  /* now test what processor family we have to choose the right event table */
  family = (signature >> 8) & 0xf;
  if (family == 5)
    {
      /* Pentium processor; test now for Pentium/MMX */
      if(((feature_flags >> 23) & 0x1) == 0)
	{
	  /* Pentium without MMX */
	  descr->event_table = event_table_pentium;
	}
      else
	/* Pentium MMX */
	descr->event_table = event_table_pentium_mmx;
    }

  else if (family == 6)
    /* Pentium Pro, Pentium II, Pentium III */
    descr->event_table = event_table_pentium2;

  else if (family == 15)
    /* Pentium IV (Willamette) */
    descr->event_table = event_table_pentium2;

  else
    /* something we can't handle yet */
    return PCL_NOT_SUPPORTED;

   /* open /dev/perfctr for read/write */
  if((hw_descr->self = vperfctr_attach()) == NULL)
    return PCL_FAILURE;

  /* everything ok */
  return PCL_SUCCESS;
}


/*                                                                            */
/*============================================================================*/
/* Starts hardware counters.                                                  */
/* max_counter (> 0) gives the maximum number of counters                     */


int
PCL_driver_start(
		 PCL_DESCR_TYPE d, /* I: descriptor */
		 int max_counter,
		 PCL_BIT_MASK_TYPE counter_used_mask,
		 PCL_DRIVER_COMMAND_TYPE *commands,
		 unsigned int count_mode
		 )

{
  int sel0, sel1, cpl;
  PCL_BIT_MASK_TYPE used_mask;
  PCL_state_descr *descr;
  HW_state_descr *hw_descr;
  struct perfctr_control control;


  /* get state descriptor */
  descr = d;
  hw_descr = &descr->hw_state_descr;

  /* set up used mask */
  used_mask = counter_used_mask & PCL_mask[max_counter-1];

  /* check count mode */
  if(count_mode == PCL_MODE_USER)
    /* count only in user mode */
    cpl = 2;
  else if(count_mode == PCL_MODE_SYSTEM)
    /* count only system mode */
    cpl = 1;
  else
    /* count in system or user mode */
    cpl = 3;


  if ((descr->event_table == event_table_pentium)
      || (descr->event_table == event_table_pentium_mmx))
    {
      /* sub-fields in the Control and Event Select Register (CESR)
       *     CC0, CC1: CPL Level to Monitor, possibilities are
       *               000 = Count Nothing
       *               001 = Count Event while CPL = 0,1,2
       *               010 = Count Event while CPL = 3
       *               011 = Count Event while CPL = 0,1,2,3
       *               100 = Count Nothing
       *               101 = Count Clocks while CPL = 0,1,2
       *               110 = Count Clocks while CPL = 3
       *               111 = Count Clocks while CPL = 0,1,2,3
       *               Here we only use 000, 001, 010 and 011.
       */
      union
      {
	struct pentium_cesr
	{
	  unsigned int es0:6;        /* event select counter 0 */
	  unsigned int cc0:3;        /* counter control 0 (see above) */
	  unsigned int pc0:1;        /* pin control bit 0
					0=event increment, 1=event overflow */
	  unsigned int reserved1:6;  /* reserved */
	  unsigned int es1:6;        /* event select counter 1 */
	  unsigned int cc1:3;        /* counter control 1 (see cc0) */
	  unsigned int pc1:1;        /* pin control bit 1 */
	  unsigned int reserved2:6;  /* reserved */
	} cesr;
	unsigned int word;           /* to initialize in one assignment */
      } cesr;
      
      /* Extract selected events */
      sel0 = sel1 = 0;
      if(IS_USED(used_mask, 0))
	sel0 = commands[0];
      if(IS_USED(used_mask, 1))
	sel1 = commands[1];

      /* set CESR */
      cesr.word = 0;
      cesr.cesr.es0 = sel0;
      cesr.cesr.es1 = sel1;
      cesr.cesr.cc0 = cesr.cesr.cc1 = cpl;
      
      /* initialize control structure */
      memset(&control, 0, sizeof control);
      control.evntsel[0] = cesr.word;
      if(vperfctr_control(hw_descr->self, &control) < 0)
	return PCL_FAILURE;

      /* read actual counter values as base values */
      if(vperfctr_read_state(hw_descr->self, &hw_descr->perfctr_state) < 0)
	return PCL_FAILURE;

      /* save base values */
      hw_descr->base_values[0] = hw_descr->perfctr_state.sum.ctr[1];
      hw_descr->base_values[1] = hw_descr->perfctr_state.sum.ctr[2];
    }


  else if (descr->event_table == event_table_pentium2)
    {
      union
      {
	struct pentium2_cesr
	{
	  unsigned int event_select:8;  /* event select */
	  unsigned int unit_mask:8;     /* further qualifies event (MESI) */
	  unsigned int usr:1;           /* count in user mode (cpl=1,2,3) */
	  unsigned int os:1;            /* count in os mode (cpl=0) */
	  unsigned int e:1;             /* edge detect */
	  unsigned int pc:1;            /* pin control */
	  unsigned int aint:1;          /* local APIC interrupt enable on
					   overflow */
	  unsigned int reserved:1;      /* reserved */
	  unsigned int en:1;            /* enable counters (sel0 only) */
	  unsigned int inv:1;           /* invert counter mask */
	  unsigned int counter_mask:8;  /* if!0, compare with events */
	} cesr;
	unsigned int word;              /* to initialize in one assignment */
      } cesr0, cesr1;


      /* set all bits to 0 */
      cesr0.word = cesr1.word = 0;

      /* additional fields */
      cesr0.cesr.usr = cesr1.cesr.usr = (cpl & 0x2) >> 1;
      cesr0.cesr.os = cesr1.cesr.os = cpl & 0x1;
      cesr0.cesr.en = 1;
      /* all other fields get 0 by the above initialization */

      if(IS_USED(used_mask, 0))
	{
	  /* events */
	  cesr0.cesr.event_select = sel0 = commands[0];
	  /* count events for all MESI states (applies only to some events) */
	  if((sel0 == 0x28) || (sel0 == 0x29)
	     || (sel0 == 0x2a) || (sel0 == 0x2e))
	    cesr0.cesr.unit_mask = 0xf;
	}
      else
	/* reset */
	cesr0.word = 0;

      if(IS_USED(used_mask, 1))
	{
	  cesr1.cesr.event_select = sel1 = commands[1];
	  if((sel1 == 0x28) || (sel1 == 0x29)
	     || (sel1 == 0x2a) || (sel1 == 0x2e))
	    cesr1.cesr.unit_mask = 0xf;
	}
      else
	/* reset */
	cesr1.word = 0;

	
      /* initialize control structure */
      memset(&control, 0, sizeof control);
      control.evntsel[0] = cesr0.word;
      control.evntsel[1] = cesr1.word;

      if(vperfctr_control(hw_descr->self, &control) < 0)
	return PCL_FAILURE;

      /* read actual counter values as base values */
      if(vperfctr_read_state(hw_descr->self, &hw_descr->perfctr_state) < 0)
	return PCL_FAILURE;

      /* save base values */
      hw_descr->base_values[0] = hw_descr->perfctr_state.sum.ctr[1];
      hw_descr->base_values[1] = hw_descr->perfctr_state.sum.ctr[2];
    }

  /* get actual clock value (to measure cycles) */
  rdtsc(&hw_descr->clock_value);

  return PCL_SUCCESS;
}


/*                                                                            */
/*============================================================================*/
/* Read hardware counters                                                     */

int
PCL_driver_read(
		PCL_DESCR_TYPE d, /* I: descriptor */
		int max_counter,
		PCL_BIT_MASK_TYPE counter_used_mask,
		PCL_CNT_TYPE *counter_values
		)

{
  long long clocks;
  PCL_BIT_MASK_TYPE used_mask;
  PCL_state_descr *descr;
  HW_state_descr *hw_descr;
  struct vperfctr_state *state;
  

  /* get state descriptor */
  descr = d;
  hw_descr = &descr->hw_state_descr;

  /* set up used mask */
  used_mask = counter_used_mask & PCL_mask[max_counter-1];

  /* get actual cycle count */
  if(IS_USED(used_mask, 2))
    {
      rdtsc(&clocks);
      counter_values[2] = clocks - hw_descr->clock_value;
    }


  /* get pointer to perfctr state */
  state = &hw_descr->perfctr_state;

  /* read counter values */
  if(vperfctr_read_state(hw_descr->self, state) < 0)
    return PCL_FAILURE;

  if(IS_USED(used_mask, 0))
    counter_values[0] = state->sum.ctr[1] - hw_descr->base_values[0];
  if(IS_USED(used_mask, 1))
    counter_values[1] = state->sum.ctr[2] - hw_descr->base_values[2];

  return PCL_SUCCESS;
}


/*                                                                            */
/*============================================================================*/
/*                                                                            */


int
PCL_driver_stop(
		PCL_DESCR_TYPE d /* I: descriptor */
		)

{
  PCL_state_descr *descr;
  HW_state_descr *hw_descr;


  /* get state descriptor */
  descr = d;
  hw_descr = &descr->hw_state_descr;

  /* stop counting */
  if(vperfctr_stop(hw_descr->self) == -1)
    return PCL_FAILURE;

  /* everything ok */
  return PCL_SUCCESS;
}


/*                                                                            */
/*============================================================================*/
/* close                                                                      */

int
PCL_driver_close(
		 PCL_DESCR_TYPE d /* I: descriptor */
		 )

{
  PCL_state_descr *descr;
  HW_state_descr *hw_descr;


  /* get state descriptor */
  descr = d;
  hw_descr = &descr->hw_state_descr;

  /* unlink driver (mmap etc.) */
  if(vperfctr_unlink(hw_descr->self) < 0)
    return PCL_FAILURE;

  /* close connection */
  vperfctr_close(hw_descr->self);

  /* everything ok */
  return PCL_SUCCESS;
}


/*============================================================================*
 *                             that's all folks                               *
 *============================================================================*/
